/* /////////////////////////////////////////////////////////////////////////////
 * File:        dlfcn.c
 *
 * Purpose:     Implementation of the dlopen, dlclose(), dlsym() and dlerror()
 *              API functions.
 *
 * Created      1st January 2004
 * Updated:     15th May 2004
 *
 * Author:      Matthew Wilson, Synesis Software Pty Ltd.
 *
 * License:     (Licensed under the Synesis Software Standard Source License)
 *
 *              Copyright (C) 2002-2004, Synesis Software Pty Ltd.
 *
 *              All rights reserved.
 *
 *              www:        http://www.synesis.com.au/
 *                          http://www.synesis.com.au/software/
 *
 *              email:      software@synesis.com.au
 *                          software@synesis-group.com
 *
 *              Redistribution and use in source and binary forms, with or
 *              without modification, are permitted provided that the following
 *              conditions are met:
 *
 *              (i) Redistributions of source code must retain the above
 *              copyright notice and contact information, this list of
 *              conditions and the following disclaimer.
 *
 *              (ii) Any derived versions of this software (howsoever modified)
 *              remain the sole property of Synesis Software.
 *
 *              (iii) Any derived versions of this software (howsoever modified)
 *              remain subject to all these conditions.
 *
 *              (iv) Neither the name of Synesis Software nor the names of any
 *              subdivisions, employees or agents of Synesis Software, nor the
 *              names of any other contributors to this software may be used to
 *              endorse or promote products derived from this software without
 *              specific prior written permission.
 *
 *              This source code is provided by Synesis Software "as is" and any
 *              warranties, whether expressed or implied, including, but not
 *              limited to, the implied warranties of merchantability and
 *              fitness for a particular purpose are disclaimed. In no event
 *              shall the Synesis Software be liable for any direct, indirect,
 *              incidental, special, exemplary, or consequential damages
 *              (including, but not limited to, procurement of substitute goods
 *              or services; loss of use, data, or profits; or business
 *              interruption) however caused and on any theory of liability,
 *              whether in contract, strict liability, or tort (including
 *              negligence or otherwise) arising in any way out of the use of
 *              this software, even if advised of the possibility of such
 *              damage.
 *
 * ////////////////////////////////////////////////////////////////////////// */


# define _SYNSOFT_VER_C_DLFCN_MAJOR     1
# define _SYNSOFT_VER_C_DLFCN_MINOR     0
# define _SYNSOFT_VER_C_DLFCN_REVISION	4
# define _SYNSOFT_VER_C_DLFCN_EDIT      4

#include <windows.h>

#include <dlfcn.h>

/* ////////////////////////////////////////////////////////////////////////// */

#ifndef WIN32
# error This file is only currently defined for Win32 compilation units
#endif /* WIN32 */

/* /////////////////////////////////////////////////////////////////////////////
 * Constants and definitions
 */

#define RTLD_LAZY			0x00001				/* Lazy function call binding. */
#define RTLD_NOW			0x00002				/* Immediate function call binding. */
#define	RTLD_BINDING_MASK   0x00003				/* Mask of binding time value. */

/* /////////////////////////////////////////////////////////////////////////////
 * Statics
 */

static const char	*s_error;

/* /////////////////////////////////////////////////////////////////////////////
 * Internal functions
 */

static void set_error_(DWORD dwErr)
{
	switch(dwErr)
	{
		case	ERROR_SUCCESS:
			s_error = NULL;
			break;
		case	ERROR_MOD_NOT_FOUND:
			s_error = "Module not found";
			break;
		case	ERROR_PROC_NOT_FOUND:
			s_error = "Symbol not found";
			break;
		case	ERROR_BAD_EXE_FORMAT:
			s_error = "Invalid image format";
			break;
		case	ERROR_SHARING_VIOLATION:
			s_error = "Sharing violation";
			break;
		default:
			s_error = "Operation failed";
			break;
	}
}

/* /////////////////////////////////////////////////////////////////////////////
 * API functions
 */

#ifdef __cplusplus
extern "C"
{
#endif /* __cplusplus */

/** 
 *
 */
void *dlopen(char const *moduleName, int mode)
{
	HMODULE	hModule;

	if(mode & RTLD_LAZY)
	{
		OutputDebugStringA("Library does not support RTLD_LAZY; using RTLD_NOW\n");
	}

	hModule =	LoadLibraryA(moduleName);

	if(NULL == hModule)
	{
		set_error_(GetLastError());
	}
	else
	{
		set_error_(ERROR_SUCCESS);
	}

	return hModule;
}

/** 
 *
 */
int dlclose(void *hModule)
{
	if(!FreeLibrary((HMODULE)hModule))
	{
		set_error_(GetLastError());

		return 1;
	}

	set_error_(ERROR_SUCCESS);

	return 0;
}

/** 
 *
 */
void *dlsym(void *hModule, char const *symbolName)
{
	void	*symbol = GetProcAddress((HMODULE)hModule, symbolName);

	if(NULL == symbol)
	{
		set_error_(GetLastError());
	}
	else
	{
		set_error_(ERROR_SUCCESS);
	}

	return symbol;
}

/** 
 *
 */
char const *dlerror(void)
{
	char const	*err = s_error;

	s_error = NULL;

	return err;
}

#ifdef __cplusplus
} // extern "C"
#endif /* __cplusplus */

/* ////////////////////////////////////////////////////////////////////////// */
